#include "generator.hpp"
#include "../include/hash.hpp"
#include <iostream>
#include <fstream>
#include <filesystem>
#include <sstream>
#include <vector>

namespace fs = std::filesystem;

static std::string sanitize(const std::string& name) {
    std::string out = name;
    for (char& c : out) if (!isalnum(c)) c = '_';
    return out;
}

void generate_file_code(const std::string& source, const std::string& destfolder, bool silent) {
    fs::path src(source);
    fs::path dest(destfolder);
    std::string ns = "recreate_" + sanitize(src.stem().string());
    std::string cppname = "_" + src.stem().string() + ".cpp";
    std::string hppname = "_" + src.stem().string() + ".hpp";
    fs::create_directories(dest);
    std::ifstream in(source, std::ios::binary);
    std::ostringstream oss;
    oss << in.rdbuf();
    std::string filedata = oss.str();
    std::string hash = hash_data(filedata);
    // Write HPP
    std::ofstream hpp(dest / hppname);
    hpp << "#pragma once\n#include <string>\nnamespace " << ns << " {\nbool recreate_file(std::string destination_folder);\n}\n";
    // Write CPP
    std::ofstream cpp(dest / cppname);
    cpp << "#include <fstream>\n#include <filesystem>\n#include <string>\n#include <iostream>\n#include <iomanip>\n#include <openssl/sha.h>\n"
        << "#include \"" << hppname << "\"\n"
        << "namespace " << ns << " {\n";
    cpp << "static const unsigned char filedata[] = {";
    for (size_t i = 0; i < filedata.size(); ++i) {
        if (i % 16 == 0) cpp << "\\n    ";
        cpp << "0x" << std::hex << std::setw(2) << std::setfill('0') << (int)(unsigned char)filedata[i];
        if (i + 1 != filedata.size()) cpp << ", ";
    }
    cpp << "\n};\n";
    cpp << "static const size_t filedata_len = " << filedata.size() << ";\n";
    cpp << "static const char* file_hash = \"" << hash << "\";\n";
    cpp << R"(
bool recreate_file(std::string destination_folder) {
    namespace fs = std::filesystem;
    fs::path outpath = fs::path(destination_folder) / "" << src.filename().string() << R"(";
    std::string existing_hash;
    if (fs::exists(outpath)) {
        std::ifstream in(outpath, std::ios::binary);
        std::ostringstream oss;
        oss << in.rdbuf();
        std::string data = oss.str();
        unsigned char hash[SHA256_DIGEST_LENGTH];
        SHA256_CTX sha256;
        SHA256_Init(&sha256);
        SHA256_Update(&sha256, data.c_str(), data.size());
        SHA256_Final(hash, &sha256);
        std::stringstream ss;
        for (int i = 0; i < SHA256_DIGEST_LENGTH; ++i)
            ss << std::hex << std::setw(2) << std::setfill('0') << (int)hash[i];
        existing_hash = ss.str();
    }
    bool needs_write = !fs::exists(outpath) || existing_hash != file_hash;
    if (needs_write) {
        std::ofstream out(outpath, std::ios::binary);
        out.write(reinterpret_cast<const char*>(filedata), filedata_len);
    }
    std::cout << "[dehydrate] " << outpath << ": ";
    if (!fs::exists(outpath)) {
        std::cout << "created\n";
    } else if (needs_write) {
        std::cout << "updated (hash changed)\n";
    } else {
        std::cout << "unchanged (hash match)\n";
    }
    return needs_write;
}
)";
    cpp << "}\n";
    if (!silent) {
        std::cout << "[dehydrate] Generated: " << (dest / cppname) << ", " << (dest / hppname) << std::endl;
    }
}

// Helper to recursively collect all files in a directory
template<typename F>
void walk_dir(const fs::path& dir, F&& f) {
    for (auto& p : fs::recursive_directory_iterator(dir)) {
        if (fs::is_regular_file(p)) f(p.path());
    }
}

void generate_folder_code(const std::string& source, const std::string& destfolder, bool silent) {
    fs::path src(source);
    fs::path dest(destfolder);
    std::string ns = "recreate_" + sanitize(src.stem().string());
    std::string cppname = "_" + src.stem().string() + ".cpp";
    std::string hppname = "_" + src.stem().string() + ".hpp";
    fs::create_directories(dest);
    // Collect all files
    std::vector<fs::path> files;
    walk_dir(src, [&](const fs::path& p) { files.push_back(p); });
    // Write HPP
    std::ofstream hpp(dest / hppname);
    hpp << "#pragma once\n#include <string>\nnamespace " << ns << " {\nbool recreate_tree(std::string destination_folder);\n}\n";
    // Write CPP
    std::ofstream cpp(dest / cppname);
    cpp << "#include <fstream>\n#include <filesystem>\n#include <string>\n#include <iostream>\n#include <iomanip>\n#include <openssl/sha.h>\n"
        << "#include \"" << hppname << "\"\n"
        << "namespace " << ns << " {\n";
    // Embed all files
    for (const auto& file : files) {
        std::ifstream in(file, std::ios::binary);
        std::ostringstream oss;
        oss << in.rdbuf();
        std::string filedata = oss.str();
        std::string hash = hash_data(filedata);
        std::string rel = fs::relative(file, src).string();
        std::string var = sanitize(rel);
        cpp << "static const unsigned char data_" << var << "[] = {";
        for (size_t i = 0; i < filedata.size(); ++i) {
            if (i % 16 == 0) cpp << "\\n    ";
            cpp << "0x" << std::hex << std::setw(2) << std::setfill('0') << (int)(unsigned char)filedata[i];
            if (i + 1 != filedata.size()) cpp << ", ";
        }
        cpp << "\n};\n";
        cpp << "static const size_t len_" << var << " = " << filedata.size() << ";\n";
        cpp << "static const char* hash_" << var << " = \"" << hash << "\";\n";
        cpp << "static const char* rel_" << var << " = \"" << rel << "\";\n";
    }
    // Write recreate_tree
    cpp << R"(
bool recreate_tree(std::string destination_folder) {
    namespace fs = std::filesystem;
    bool any_written = false;
)";
    for (const auto& file : files) {
        std::string rel = fs::relative(file, src).string();
        std::string var = sanitize(rel);
        cpp << "    {\n        fs::path outpath = fs::path(destination_folder) / rel_" << var << ";\n"
            << "        fs::create_directories(outpath.parent_path());\n"
            << "        std::string existing_hash;\n"
            << "        if (fs::exists(outpath)) {\n"
            << "            std::ifstream in(outpath, std::ios::binary);\n"
            << "            std::ostringstream oss; oss << in.rdbuf();\n"
            << "            std::string data = oss.str();\n"
            << "            unsigned char hash[SHA256_DIGEST_LENGTH];\n"
            << "            SHA256_CTX sha256; SHA256_Init(&sha256);\n"
            << "            SHA256_Update(&sha256, data.c_str(), data.size());\n"
            << "            SHA256_Final(hash, &sha256);\n"
            << "            std::stringstream ss;\n"
            << "            for (int i = 0; i < SHA256_DIGEST_LENGTH; ++i)\n"
            << "                ss << std::hex << std::setw(2) << std::setfill('0') << (int)hash[i];\n"
            << "            existing_hash = ss.str();\n"
            << "        }\n"
            << "        bool needs_write = !fs::exists(outpath) || existing_hash != hash_" << var << ";\n"
            << "        if (needs_write) {\n"
            << "            std::ofstream out(outpath, std::ios::binary);\n"
            << "            out.write(reinterpret_cast<const char*>(data_" << var << "), len_" << var << ");\n"
            << "        }\n"
            << "        std::cout << \"[dehydrate] \" << outpath << ": ";\n"
            << "        if (!fs::exists(outpath)) {\n"
            << "            std::cout << \"created\\n\";\n"
            << "        } else if (needs_write) {\n"
            << "            std::cout << \"updated (hash changed)\\n\";\n"
            << "        } else {\n"
            << "            std::cout << \"unchanged (hash match)\\n\";\n"
            << "        }\n"
            << "        any_written = any_written || needs_write;\n"
            << "    }\n";
    }
    cpp << "    return any_written;\n}\n";
    cpp << "}\n";
    if (!silent) {
        std::cout << "[dehydrate] Generated: " << (dest / cppname) << ", " << (dest / hppname) << std::endl;
    }
} 